from sanic.views import HTTPMethodView
from sanic.response import json
from naas.runner.env_var import n_env

from pathlib import Path


def get_folder_size(folder):
    return ByteSize(sum(file.stat().st_size for file in Path(folder).rglob("*")))


class ByteSize(int):

    _kB = 1024
    _suffixes = "B", "kB", "MB", "GB", "PB"

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        self.bytes = self.B = int(self)
        self.kilobytes = self.kB = self / self._kB ** 1
        self.megabytes = self.MB = self / self._kB ** 2
        self.gigabytes = self.GB = self / self._kB ** 3
        self.petabytes = self.PB = self / self._kB ** 4
        *suffixes, last = self._suffixes
        suffix = next(
            (suffix for suffix in suffixes if 1 < getattr(self, suffix) < self._kB),
            last,
        )
        self.readable = suffix, getattr(self, suffix)

        super().__init__()

    def __str__(self):
        return self.__format__(".2f")

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, super().__repr__())

    def __format__(self, format_spec):
        suffix, val = self.readable
        return "{val:{fmt}} {suf}".format(val=val, fmt=format_spec, suf=suffix)

    def __sub__(self, other):
        return self.__class__(super().__sub__(other))

    def __add__(self, other):
        return self.__class__(super().__add__(other))

    def __mul__(self, other):
        return self.__class__(super().__mul__(other))

    def __rsub__(self, other):
        return self.__class__(super().__sub__(other))

    def __radd__(self, other):
        return self.__class__(super().__add__(other))

    def __rmul__(self, other):
        return self.__class__(super().__rmul__(other))


class PerformanceController(HTTPMethodView):
    def __init__(self, *args, **kwargs):
        super(PerformanceController, self).__init__(*args, **kwargs)

    async def get(self, request, mode=""):
        modes = {
            "cpu": PerformanceController.getCpu,
            "ram": PerformanceController.getRam,
            "storage": PerformanceController.getStorage,
        }
        if modes.get(mode) is not None:
            return json({mode: modes[mode](self)})
        else:
            perf = {}
            for key, value in modes.items():
                perf[key] = value(self)
            return json(perf)

    # TODO make the functions to get the cpu and ram value
    def getCpu(self):
        return "/"

    def getRam(self):
        return "/"

    def getStorage(self):
        return str(get_folder_size(n_env.server_root))



from .ntypes import t_scheduler, t_output, t_add, t_update, t_delete
from .manager import Manager
import pandas as pd
import pretty_cron
import requests
import pycron


class Scheduler:
    naas = None
    role = t_scheduler
    manager = None

    def __init__(self):
        self.manager = Manager(t_scheduler)
        self.path = self.manager.path

    def list(self, path=None):
        return self.manager.list_prod("list_history", path)

    def list_output(self, path=None):
        return self.manager.list_prod("list_output", path)

    def get(self, path=None, histo=None):
        return self.manager.get_file(path, histo=histo)

    def get_output(self, path=None, histo=None):
        return self.manager.get_file(path, t_output, histo)

    def clear(self, path=None, histo=None):
        return self.manager.clear_file(path, None, histo)

    def clear_output(self, path=None, histo=None):
        return self.manager.clear_file(path, t_output, histo)

    def status(self):
        req = requests.get(url=f"{self.manager.naas_api}/scheduler")
        req.raise_for_status()
        jsn = req.json()
        print(jsn)
        return jsn

    def pause(self):
        req = requests.get(url=f"{self.manager.naas_api}/scheduler/pause")
        req.raise_for_status()
        jsn = req.json()
        print(jsn)
        return jsn

    def resume(self):
        req = requests.get(url=f"{self.manager.naas_api}/scheduler/resume")
        req.raise_for_status()
        jsn = req.json()
        print(jsn)
        return jsn

    def currents(self, raw=False):
        json_data = self.manager.get_naas()
        json_filtered = []
        for item in json_data:
            if item["type"] == self.role and item["status"] != t_delete:
                if raw:
                    json_filtered.append(item)
                else:
                    json_filtered.append({"path": item["path"], "value": item["value"]})
        if raw is False:
            df = pd.DataFrame(json_filtered)
            df = df.style.format({"value": pretty_cron.prettify_cron})
            return df
        return json_filtered

    def __check_cron(self, text):
        res = False
        try:
            pycron.is_now(text)
            res = True
        except ValueError:
            pass
        return res

    def add(self, path=None, recurrence=None, cron=None, params={}, debug=False):
        if self.manager.is_production():
            print("No add done, you are in production\n")
            return
        if recurrence:
            print("recurrence is deprecated use cron arg instead")
        cron = recurrence if recurrence else cron
        if not cron:
            print("No cron provided\n")
            return
        if not self.__check_cron(cron):
            print(f"WARNING : Recurrence wrong format {cron}")
            return
        current_file = self.manager.get_path(path)
        status = t_add
        try:
            self.manager.get_value(current_file, False)
            status = t_update
        except:  # noqa: E722
            pass
        self.manager.add_prod(
            {
                "type": self.role,
                "path": current_file,
                "status": status,
                "params": params,
                "value": cron,
            },
            debug,
        )
        cron_string = pretty_cron.prettify_cron(cron)
        print("👌 Well done! Your Notebook has been sent to production. \n")
        print(
            f'⏰ It will be scheduled "{cron_string}" (more on the syntax on https://crontab.guru/).\n'
        )
        print('Ps: to remove the "Scheduler", just replace .add by .delete')

    def delete(self, path=None, all=True, debug=False):
        if self.manager.is_production():
            print("No delete done, you are in production\n")
            return
        current_file = self.manager.get_path(path)
        self.manager.del_prod({"type": self.role, "path": current_file}, debug)
        print("🗑 Done! Your Scheduler has been remove from production.\n")
        if all is True:
            self.clear(current_file, "all")
            self.clear_output(current_file, "all")



from naas.ntypes import (
    t_scheduler,
    t_start,
    t_main,
    t_health,
    t_error,
    t_busy,
    t_delete,
)
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from naas.callback import Callback
import apscheduler.schedulers.base
from .env_var import n_env
import traceback
import requests
import datetime
import asyncio
import aiohttp
import pycron
import time
import pytz
import uuid
import os


async def fetch(url):
    with aiohttp.ClientSession() as session:
        async with session.get(url, timeout=n_env.scheduler_timeout) as response:
            return await response.json()


class Scheduler:
    __scheduler = None
    __nb = None
    __logger = None
    __jobs = None

    def __init__(self, logger, jobs, notebooks, loop):
        self.__logger = logger
        self.__nb = notebooks
        self.__jobs = jobs
        self.__scheduler = AsyncIOScheduler({"event_loop": loop})
        self.state = self.__scheduler.state

    def status(self):
        if self.__scheduler.state == apscheduler.schedulers.base.STATE_RUNNING:
            return "running"
        elif self.__scheduler.state == apscheduler.schedulers.base.STATE_PAUSED:
            return "paused"

    def stop(self):
        print("Stop scheduler")
        if self.__scheduler is not None:
            self.__scheduler.pause()
            self.__scheduler.remove_job(n_env.scheduler_job_name)
            self.__scheduler.shutdown(wait=False)
            self.__scheduler = None

    async def start(self, test_mode=False):
        if test_mode:
            await self.__scheduler_function()
        else:
            if self.__scheduler.state != apscheduler.schedulers.base.STATE_RUNNING:
                self.__scheduler.add_job(
                    func=self.__scheduler_function,
                    trigger="interval",
                    id=n_env.scheduler_job_name,
                    seconds=n_env.scheduler_interval,
                    max_instances=n_env.scheduler_job_max,
                )
                self.__scheduler.start()
                uid = str(uuid.uuid4())
                self.__logger.info(
                    {
                        "id": uid,
                        "type": t_main,
                        "filepath": "sheduler",
                        "status": f"start SCHEDULER seconds={n_env.scheduler_interval}, max_instances={n_env.scheduler_job_max}",
                    }
                )

    async def __check_run(self, uid, file_filepath, current_type, last_update_str):
        running = await self.__jobs.is_running(uid, file_filepath, current_type)
        if last_update_str and running:
            try:
                last_update = datetime.datetime.strptime(
                    last_update_str, "%d/%m/%y %H:%M:%S"
                )
                # Timeout run 1h
                timeout_date = datetime.datetime.today() - datetime.timedelta(hours=1)
                running = True if last_update > timeout_date else False
            except ValueError:
                pass
        return running

    async def __scheduler_greenlet(self, main_uid, current_time, job):
        value = job.get("value")
        current_type = job.get("type")
        file_filepath = job.get("path")
        last_update = job.get("lastUpdate")
        status = job.get("status")
        params = job.get("params", dict())
        uid = str(uuid.uuid4())
        try:
            running = await self.__check_run(
                uid, file_filepath, current_type, last_update
            )
            if (
                current_type == t_scheduler
                and value is not None
                and pycron.is_now(value, current_time)
                and status != t_delete
                and not running
            ):
                self.__logger.info(
                    {
                        "main_id": str(main_uid),
                        "id": uid,
                        "type": t_scheduler,
                        "status": t_start,
                        "filepath": file_filepath,
                    }
                )
                await self.__jobs.update(
                    uid, file_filepath, t_scheduler, value, params, t_start
                )
                res = await self.__nb.exec(uid, job.copy())
                if res.get("error"):
                    self.__logger.error(
                        {
                            "main_id": str(main_uid),
                            "id": uid,
                            "type": t_scheduler,
                            "status": t_error,
                            "filepath": file_filepath,
                            "duration": res.get("duration"),
                            "error": str(res.get("error")),
                        }
                    )
                    await self.__jobs.update(
                        uid,
                        file_filepath,
                        t_scheduler,
                        value,
                        params,
                        t_error,
                        res.get("duration"),
                    )
                    return
                next_url = params.get("next_url", None)
                if next_url is not None and "https://" not in next_url:
                    self.__logger.error(
                        {
                            "id": uid,
                            "type": t_scheduler,
                            "status": "next_url",
                            "filepath": file_filepath,
                            "url": next_url,
                            "error": "url not in right format",
                        }
                    )
                elif next_url is not None and "https://" in next_url:
                    self.__logger.info(
                        {
                            "id": uid,
                            "type": t_scheduler,
                            "status": "next_url",
                            "filepath": file_filepath,
                            "url": next_url,
                        }
                    )
                    try:
                        await fetch(url=next_url)
                    except:  # noqa: E722
                        tb = traceback.format_exc()
                        self.__logger.error(
                            {
                                "id": main_uid,
                                "type": t_scheduler,
                                "status": t_error,
                                "filepath": file_filepath,
                                "error": "Error in next_url",
                                "traceback": str(tb),
                            }
                        )
                self.__logger.info(
                    {
                        "main_id": str(main_uid),
                        "id": uid,
                        "type": t_scheduler,
                        "status": t_health,
                        "filepath": file_filepath,
                        "duration": res.get("duration"),
                    }
                )
                await self.__jobs.update(
                    uid,
                    file_filepath,
                    t_scheduler,
                    value,
                    params,
                    t_health,
                    res.get("duration"),
                )
            elif running:
                self.__logger.info(
                    {
                        "main_id": str(main_uid),
                        "id": uid,
                        "type": t_scheduler,
                        "status": t_busy,
                        "filepath": file_filepath,
                    }
                )
        except:  # noqa: E722
            tb = traceback.format_exc()
            self.__logger.error(
                {
                    "id": main_uid,
                    "type": t_scheduler,
                    "status": t_error,
                    "filepath": file_filepath,
                    "error": "Unknow error",
                    "traceback": str(tb),
                }
            )
            await self.__jobs.update(
                uid,
                file_filepath,
                t_scheduler,
                value,
                params,
                t_error,
            )

    def getTerminals(self):
        try:
            r = requests.get(
                f"{n_env.user_url}/api/terminals",
                headers={
                    "Authorization": f"token {n_env.token}",
                },
            )
            r.raise_for_status()
            return r.json()
        except Exception as e:
            print(e)
            return {}

    def getSessions(self):
        try:
            r = requests.get(
                f"{n_env.user_url}/api/sessions",
                headers={
                    "Authorization": f"token {n_env.token}",
                },
            )
            r.raise_for_status()
            return r.json()
        except Exception as e:
            print(e)
            return {}

    async def analytics(self, uid):
        try:
            curdate = datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")
            data = {
                "date": curdate,
                "gpu": os.environ.get("NAAS_GPU", "NO"),
                "jobs": await self.__jobs.list(uid),
                "kernels": self.getSessions(),
                "terminals": self.getTerminals(),
            }
            if n_env.report_callback:
                Callback().add(
                    auto_delete=False,
                    uuid=f"naas_analytics__{curdate}",
                    default_result=data,
                    no_override=True,
                )
        except Exception as e:
            tb = traceback.format_exc()
            self.__logger.error(
                {
                    "id": uid,
                    "type": t_scheduler,
                    "filepath": "analytics",
                    "status": t_error,
                    "error": str(e),
                    "traceback": tb,
                }
            )

    async def __scheduler_function(self):
        # Create unique for scheduling step (one step every minute)
        main_uid = str(uuid.uuid4())
        all_start_time = time.time()
        try:
            if n_env.scheduler_interval == 60:
                current_time = datetime.datetime.now(tz=pytz.timezone(n_env.tz))
            elif n_env.scheduler_interval == 1:
                # for speed test in ci
                current_time = datetime.datetime.now(tz=pytz.timezone(n_env.tz))
                current_sec = current_time.second
                current_time = current_time.replace(minute=current_sec)
            else:
                raise ValueError(
                    f"naas doesn't support NAAS_SCHEDULER_TIME={n_env.scheduler_interval}"
                )
            print(f"\n\n================ {current_time} ============== \n\n")
            self.__logger.info(
                {
                    "id": main_uid,
                    "type": t_scheduler,
                    "filepath": "scheduler",
                    "status": t_start,
                }
            )
            jobs = await self.__jobs.list(main_uid, prodPath=True)
            await asyncio.gather(
                *[
                    self.__scheduler_greenlet(main_uid, current_time, job)
                    for job in jobs
                ],
                return_exceptions=False,
            )
            duration_total = time.time() - all_start_time
            self.__logger.info(
                {
                    "id": main_uid,
                    "type": t_scheduler,
                    "filepath": "scheduler",
                    "status": t_health,
                    "duration": duration_total,
                }
            )
            # We disable this for now as it is not used and is therefore filling up the database.
            # await self.analytics(main_uid)
        except Exception as e:
            tb = traceback.format_exc()
            duration_total = time.time() - all_start_time
            self.__logger.error(
                {
                    "id": main_uid,
                    "type": t_scheduler,
                    "status": t_error,
                    "filepath": "scheduler",
                    "duration": duration_total,
                    "error": str(e),
                    "traceback": tb,
                }
            )
        except:  # noqa: E722
            tb = traceback.format_exc()
            duration_total = time.time() - all_start_time
            self.__logger.error(
                {
                    "id": main_uid,
                    "type": t_scheduler,
                    "status": t_error,
                    "filepath": "scheduler",
                    "duration": duration_total,
                    "error": "Unknow error",
                    "traceback": tb,
                }
            )
